import json
import random
from pathlib import Path
from typing import Union

from storage import UploadedEntity

import portal.tools.worker_client
from portal import logger, tools
from portal.database import Project, Submission, SubmissionState, Worker
from portal.service import general

log = logger.get_logger(__name__)


class SubmissionProcessor:
    def __init__(self, submission: Submission, params: dict = None):
        self._submission = submission
        self._params = params

    @property
    def submission(self) -> Submission:
        return self._submission

    @property
    def params(self) -> dict:
        return self.submission.parameters

    @property
    def project(self) -> Project:
        return self.submission.project

    @property
    def celery(self):
        from portal.async_celery import celery_app
        return celery_app

    @property
    def storage(self):
        from portal import storage
        return storage

    def reset_task_id(self, state=None):
        if state is not None:
            self.submission.state = state

        self.submission.async_task_id = None
        general.write_entity(self.submission)

    def submission_enqueue_ended(self):
        log.info(f"[ASYNC] Submission enqueue ended: {self.submission}")
        self.reset_task_id(state=SubmissionState.QUEUED)

    def get_delay_for_submission(self):
        log.debug(f"[ASYNC] Submission delay: {self.submission}")
        time_wait = self.project.config.submissions_cancellation_period
        return time_wait

    def dispatch_submission_processing(self):
        delay = self.get_delay_for_submission()
        args = (self.submission.id, self.params)
        from .tasks import start_processing_submission
        start_processing_submission.apply_async(args=args, countdown=delay)

    def submission_store_ended(self, version: str):
        log.info(f"[ASYNC] Submission preparation ended: {self.submission}")
        self.submission.source_hash = version
        self.reset_task_id(state=SubmissionState.READY)

    def download_submission(self):
        file_params = self.params['file_params']
        log.info(f"[ASYNC] Uploading submission: {self.submission} with {file_params}")
        updated_entity: UploadedEntity = self.storage. \
            submissions.create(entity_id=self.submission.id, **file_params)
        self.submission_store_ended(version=updated_entity.version)

    def clone(self, target):
        log.info(f"[ASYNC] Cloning submission: {self.submission} to {target}")
        self.storage.submissions.clone(self.submission.id, target.id)
        self.submission_store_ended(version=self.submission.source_hash)

    def send_to_worker(self):
        # TODO: implement processing
        log.info(f"[ASYNC] Sending submission to worker: {self.submission}")
        worker = self.schedule_submission_to_worker()
        self.execute_submission(worker)

    def upload_result(self, path, file_params):
        log.info(f"[ASYNC] Uploading result for the submission "
                 f"{self.submission.id} with {file_params}")
        self.storage.results.create(entity_id=self.submission.id, **file_params)
        Path(path).unlink()
        self.reset_task_id(SubmissionState.FINISHED)

    def process_submission(self):
        log.info(f"[ASYNC] Processing submission: {self.submission}")
        self.download_submission()
        self.dispatch_submission_processing()

    def revoke_task(self):
        log.info(f'[ASYNC] Submission processing cancelled {self.submission}')
        task_id = self.submission.async_task_id
        if task_id:
            self.celery.control.revoke(task_id=task_id, terminate=True)
            self.reset_task_id(state=SubmissionState.CANCELLED)
        # TODO: Storage clean up

    # TODO implement - @mdujava
    # STUB: Select initialized worker
    def schedule_submission_to_worker(self) -> Worker:
        """Based on the features (worker tags) and preferences in project config
        schedule submission for the execution on initialized worker

        Returns(Worker): Worker instance on which the submission will be executed
        """
        workers = Worker.query.filter(Worker.is_initialized).all()
        if not workers:
            self._worker_not_available()
        worker = random.choice(workers)  # randomly select a worker
        return worker

    def execute_submission(self, worker: Worker):
        worker_client = tools.worker_client.WorkerClient(worker=worker)
        worker_client.execute_submission(self.submission)

    def _worker_not_available(self):
        log.warning(f"[PROC] Worker is no available for submission: {self.submission}")
        pass
