import flask
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal import logger, storage
from portal.rest import rest_helpers
from portal.rest.schemas import review_schema, submission_schema, submission_state_schema
from portal.service import auth, general, permissions
from portal.service.reviews import ReviewService
from portal.service.submissions import SubmissionsService

submissions_namespace = Namespace('submissions')

log = logger.get_logger(__name__)


@submissions_namespace.route('/<string:sid>')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResource(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission found', model=submission_schema)
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        # Authorization
        permissions.PermissionsService(course=course).require.read_submission(submission)
        return submission_schema.dump(submission)

    @jwt_required
    @submissions_namespace.response(204, 'Submission deleted')
    def delete(self, sid: str):
        permissions.PermissionsService().require.sysadmin()
        submission = general.find_submission(sid)
        SubmissionsService(submission=submission).delete_submission()
        return '', 204


@submissions_namespace.route('/<string:sid>/state')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionState(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission state', model=submission_state_schema)
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission(submission)
        return submission_state_schema.dump(submission)[0]

    @jwt_required
    @submissions_namespace.response(204, 'Submission state updated')
    def put(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        permissions.PermissionsService().require.sysadmin_or_self(submission.user.id)
        data = rest_helpers.parse_request_data(
            submission_state_schema, action='update_state', resource='submission'
        )
        service = SubmissionsService(submission=submission)
        service.update_submission_state(client=client, data=data)
        return '', 204


@submissions_namespace.route('/<string:sid>/result')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResult(Resource):
    @jwt_required
    def put(self, sid: str):
        raise NotImplementedError()


@submissions_namespace.route('/<string:sid>/files')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        raise NotImplementedError()


@submissions_namespace.route('/<string:sid>/files/sources/tree')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionSourcesTree(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission(submission)
        service = SubmissionsService(submission=submission)
        return service.send_files_tree()


@submissions_namespace.route('/<string:sid>/files/sources')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionSourceFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission(submission)
        service = SubmissionsService(submission=submission)
        return service.send_file_or_zip()


@submissions_namespace.route('/<string:sid>/files/test_files/tree')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionTestFilesTree(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission_group(submission)
        storage_entity = storage.test_files.get(submission.project.id)
        service = SubmissionsService(submission=submission)
        return service.send_files_tree(storage_entity)


@submissions_namespace.route('/<string:sid>/files/test_files')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionTestFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission_group(submission)
        storage_entity = storage.test_files.get(submission.project.id)
        service = SubmissionsService(submission=submission)
        return service.send_file_or_zip(storage_entity)


@submissions_namespace.route('/<string:sid>/files/results/tree')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResultFilesTree(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission_group(submission)
        storage_entity = storage.results.get(submission.id)
        service = SubmissionsService(submission=submission)
        tree = service.send_files_tree(storage_entity)
        return flask.jsonify(tree), 200


@submissions_namespace.route('/<string:sid>/files/results')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResultFiles(Resource):
    @jwt_required
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission_group(submission)
        storage_entity = storage.results.get(submission.id)
        service = SubmissionsService(submission=submission)
        return service.send_file_or_zip(storage_entity)

    @jwt_required
    def post(self, sid: str):
        submission = general.find_submission(sid)
        # authorization
        course = submission.project.course
        permissions.PermissionsService(course=course).require.client(['evaluate_submissions'])
        # todo: authorize worker
        service = SubmissionsService(submission=submission)
        task = service.upload_results_to_storage()
        return {'new_task': task.id}


@submissions_namespace.route('/<string:sid>/resubmit')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResubmit(Resource):
    @jwt_required
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        source_submission = general.find_submission(sid)
        # authorization
        course = source_submission.project.course
        perm_service = permissions.PermissionsService(course=course)
        perm_service.require.client(permissions=['resubmit_submissions'])

        data = rest_helpers.parse_request_data(
            submission_schema, action='resubmit', resource='submission'
        )
        # create new submission by copying files from the source submission in
        # storage
        service = SubmissionsService(submission=source_submission)
        new_submission = service.copy_submission(note=data['note'])
        return submission_schema.dump(new_submission), 201


@submissions_namespace.route('/<string:sid>/cancel')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
@submissions_namespace.response(204, 'Submissions canceled')
class SubmissionCancel(Resource):
    @jwt_required
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        submission = general.find_submission(sid)
        # authorization
        course = submission.project.course
        perm_service = permissions.PermissionsService(course=course)
        perm_service.require.sysadmin_or_self(submission.user.id)
        # create new submission by copying files from the source submission in
        # storage
        service = SubmissionsService(submission=submission)
        service.cancel_submission()
        return '', 204


@submissions_namespace.route('/<string:sid>/review')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionReview(Resource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submissions review', model=review_schema)
    def get(self, sid: str):
        submission = general.find_submission(sid)
        course = submission.project.course
        permissions.PermissionsService(course=course).require.read_submission(submission)
        return review_schema.dump(submission.review)[0]

    @jwt_required
    # @submissions_namespace.response(201, 'Submissions review created', model=review_schema)
    def post(self, sid: str):
        client = auth.find_client()
        submission = general.find_submission(sid)
        course = submission.project.course
        perm_service = permissions.PermissionsService(course=course)
        perm_service.require.write_review_for_submission(submission)
        data = rest_helpers.parse_request_data(
            review_schema, action='create', resource='review')

        review_service = ReviewService(submission=submission)
        if not submission.review:
            review_service.create()

        # also writes to db
        review_service.create_review_items(items=data['review_items'], author=client)

        return review_schema.dump(submission.review)[0], 201
