from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal import logger
from portal.rest import rest_helpers
from portal.rest.schemas import secret_schema, secrets_schema, worker_schema, \
    workers_schema
from portal.service import general, permissions
from portal.service.secrets import create_secret, delete_secret
from portal.service.workers import WorkerService

workers_namespace = Namespace('workers')
log = logger.get_logger(__name__)


@workers_namespace.route('')
class WorkerList(Resource):
    @jwt_required
    # @workers_namespace.response(200, 'List of workers', model=workers_schema)
    @workers_namespace.response(403, 'Not allowed to list workers')
    def get(self):
        permissions.PermissionsService().require.sysadmin()
        workers_list = WorkerService().find_all_workers()
        return workers_schema.dump(workers_list)[0]

    @jwt_required
    # @workers_namespace.response(201, 'Created worker', model=worker_schema)
    @workers_namespace.response(403, 'Not allowed to add worker')
    def post(self):
        permissions.PermissionsService().require.sysadmin()
        data = rest_helpers.parse_request_data(
            worker_schema, action='create', resource='worker'
        )
        new_worker = WorkerService().create_worker(**data)
        return worker_schema.dump(new_worker)[0], 201


@workers_namespace.route('/<string:wid>')
@workers_namespace.param('wid', 'Worker id')
@workers_namespace.response(404, 'Worker not found')
class WorkerResource(Resource):

    @jwt_required
    # @workers_namespace.response(200, 'Worker', model=worker_schema)
    @workers_namespace.response(403, 'Not allowed to access worker')
    def get(self, wid: str):
        permissions.PermissionsService().require.sysadmin_or_self(wid)
        worker = general.find_worker(wid)
        return worker_schema.dump(worker)[0]

    @jwt_required
    @workers_namespace.response(204, 'Worker deleted')
    @workers_namespace.response(403, 'Not allowed to delete worker')
    def delete(self, wid: str):
        permissions.PermissionsService().require.sysadmin()
        worker = general.find_worker(wid)
        WorkerService(worker=worker).delete_worker()
        return '', 204

    @jwt_required
    @workers_namespace.response(204, 'Worker updated')
    @workers_namespace.response(403, 'Not allowed to update worker')
    def put(self, wid: str):
        permissions.PermissionsService().require.sysadmin_or_self(wid)

        data = rest_helpers.parse_request_data(
            worker_schema, action='update', resource='worker', partial=True
        )

        worker = general.find_worker(wid)
        WorkerService(worker=worker).update_worker(data)
        return '', 204


@workers_namespace.route('/<string:wid>/secrets')
@workers_namespace.param('wid', "Worker's id")
@workers_namespace.response(404, 'Worker not found')
class WorkerSecretsController(Resource):
    @jwt_required
    @workers_namespace.response(200, 'List worker secrets')
    def get(self, wid: str):
        permissions.PermissionsService().require.sysadmin()
        worker = general.find_worker(wid)
        return secrets_schema.dump(worker.secrets)

    @jwt_required
    # @workers_namespace.response(201, 'Created worker secret', model=secret_schema)
    def post(self, wid: str):
        permissions.PermissionsService().require.sysadmin()
        data = rest_helpers.parse_request_data(
            secret_schema, action='create', resource='secret')
        worker = general.find_worker(wid)
        new_secret, value = create_secret(worker, **data)
        return {'id': new_secret.id, 'value': value}, 201


@workers_namespace.route('/<string:wid>/secrets/<string:sid>')
@workers_namespace.param('wid', "Worker's id")
@workers_namespace.param('sid', "Secret id")
@workers_namespace.response(404, 'Worker not found')
class WorkerSecretController(Resource):
    @jwt_required
    @workers_namespace.response(200, 'Worker secret detail')
    def get(self, wid: str, sid: str):
        permissions.PermissionsService().require.sysadmin()
        worker = general.find_worker(wid)
        secret = general.find_secret(worker, sid)
        return secret_schema.dump(secret)

    @jwt_required
    @workers_namespace.response(204, 'Worker secret deleted')
    def delete(self, wid: str, sid: str):
        permissions.PermissionsService().require.sysadmin()
        worker = general.find_worker(wid)
        delete_secret(worker, sid)
        return '', 204

    @jwt_required
    @workers_namespace.response(204, 'Worker secret updated')
    @workers_namespace.response(403, 'Not allowed to update worker secret')
    def put(self, wid: str, sid: str):
        permissions.PermissionsService().require.sysadmin()

        data = rest_helpers.parse_request_data(
            secret_schema, action='update', resource='secret', partial=True
        )

        worker = general.find_worker(wid)
        secret = general.find_secret(worker, sid)
        general.update_entity(secret, data, ['name', 'expires_at'])

        return '', 204
