"""
Permissions service
"""

import logging
from typing import Union

from portal.database.models import ClientType, Course, Role, User, Worker
from portal.service import auth, general
from portal.service.errors import ForbiddenError

FILTER_PERMISSION_ATTRS = ['id', 'role_id', 'role',
                           'description', 'created_at', 'updated_at']

log = logging.getLogger(__name__)


def _resolve_permissions(permissions_required_all: list, effective_permissions: dict):
    if isinstance(permissions_required_all, str):
        permissions_required_all = [permissions_required_all]
    return all(
        effective_permissions.get(permission, False) for permission in permissions_required_all)


class PermissionServiceCheck:
    def __init__(self, service: 'PermissionsService'):
        self.service = service

    def sysadmin_or_self(self, eid):
        return self.any_check(self.sysadmin(), self.service.client.is_self(eid))

    def sysadmin(self) -> bool:
        if self.service.client.type == ClientType.USER:
            return self.service.client.is_admin
        return False

    def client(self, permissions) -> bool:
        log.debug(
            f"[PERM] Client {self.service.client.id} in course "
            f"{self.service.course.codename}: {permissions}")
        if self.sysadmin():
            return True

        effective_permissions = self.service.effective_permissions_for_course()
        if not effective_permissions:
            return False

        res = any(_resolve_permissions(conjunction, effective_permissions) for conjunction in
                  permissions)

        log.debug(f"[PERM] {permissions} for {self.service.client.id}: {res}")
        return res

    def any_check(self, *params) -> bool:
        return any(params)


class PermissionServiceRequire:
    def __init__(self, service: 'PermissionsService'):
        self.service = service

    def sysadmin_or_self(self, eid):
        self.any_check(self.service.check.sysadmin_or_self(eid))

    def sysadmin(self):
        self.any_check(self.service.check.sysadmin())

    def client(self, permissions):
        self.any_check(self.service.check.client(permissions))

    def any_check(self, *checks):
        if not self.service.check.any_check(*checks):
            raise ForbiddenError(self.service.client)

    def update_course(self):
        self.client(['update_course'])

    def write_projects(self):
        self.client(['update_course', 'write_projects'])

    def write_roles(self):
        self.client(['update_course', 'write_roles'])

    def write_groups(self):
        self.client(['update_course', 'write_groups'])

    def view_course(self):
        self.client(['view_course_full', 'view_course_limited'])

    def belongs_to_group(self, group):
        checks = [
            self.service.check.client(['view_course_full']),
            (self.service.check.client(['view_course_limited'])
             and self.service.client_owner in group.users)
        ]
        self.any_check(*checks)

    def belongs_to_role(self, role: Role):
        checks = [
            self.service.check.client(['view_course_full']),
            (self.service.check.client(['view_course_limited'])
             and self.service.client_owner in role.clients)
        ]
        self.any_check(*checks)

    def read_submission(self, submission):
        checks = [
            self.service.check.client(['read_submissions_all']),
            self.service.submission_access_group(submission, ['read_submissions_groups']),
            (self.service.check.client(['read_submissions_own']) and
             submission.user == self.service.client_owner)
        ]
        self.any_check(*checks)

    def read_submission_group(self, submission):
        checks = [
            self.service.check.client(['read_submissions_all']),
            self.service.submission_access_group(submission, ['read_submissions_groups'])
        ]
        self.any_check(*checks)

    def write_review_for_submission(self, submission):
        checks = [
            self.service.check.client(['write_reviews_all']),
            self.service.submission_access_group(submission, ['write_reviews_group']),
            (self.service.check.client(['write_reviews_own']) and
             submission.user == self.service.client_owner)
        ]
        self.any_check(*checks)


class PermissionsService:
    def __init__(self, client=None, course=None):
        self._client = client or auth.find_client()
        self._course = course
        self._requires = PermissionServiceRequire(self)
        self._checks = PermissionServiceCheck(self)

    @property
    def require(self):
        return self._requires

    @property
    def check(self):
        return self._checks

    @property
    def client(self):
        return self._client

    @property
    def client_owner(self) -> Union[Worker, User]:
        return general.find_client_owner(self.client)

    @property
    def course(self) -> Course:
        return self._course

    def get_effective_permissions(self, course_id: str = None) -> dict:
        """Gets effective permissions for a client in course.
        If no course is specified, returns the client's
        effective permissions in all courses he is a part of.

        Args:
            course_id(str): Course ID

        Returns(dict): Effective permissions for the client. Keys are course IDs, values
        dictionaries of permissions with their values.
        """
        course = general.find_course(course_id, throws=False)
        courses = [course] if course else self.client.courses
        return self.build_effective_permissions(*courses)

    def build_effective_permissions(self, *courses) -> dict:
        """Builds effective permissions in a list of courses for the client

        Args:
            *courses: List of courses

        Returns(dict): Effective permissions
        """
        result = {}
        for course in courses:
            result[course.id] = self.effective_permissions_for_course(course)
        return result

    def effective_permissions_for_course(self, course: Course = None) -> dict:
        """Extracts effective permissions for a client in one course

        Returns(dict): Effective permissions dictionary
        """
        course = course or self.course
        permissions = self.client.get_permissions_for_course(course)
        result = {}
        for permission in permissions:
            for key, value in vars(permission).items():
                if not key.startswith("_") and key not in FILTER_PERMISSION_ATTRS:
                    result[key] = result.get(key) or value
        log.debug(
            f"[PERM] Effective permissions: {self.client.id} "
            f"in course {course.codename}: {result}")
        return result

    def submission_access_group(self, submission, perm):
        if self.check.client(perm):
            group_intersection = [
                group for group in self.client.groups if submission.user in group.users]
            return any(
                group in submission.project.groups for group in group_intersection)
        return False
