"""
Submissions service
"""

import logging
from pathlib import Path
from typing import Union

import flask
from celery.result import AsyncResult
from storage import entities
from werkzeug.utils import secure_filename

from portal import storage
from portal.async_celery import submission_processor, tasks
from portal.database.models import Project, Submission, SubmissionState, User, Worker
from portal.service import errors
from portal.service.general import delete_entity, write_entity

log = logging.getLogger(__name__)


def upload_files_to_storage(file):
    filename = secure_filename(file.filename)
    uploads = storage.results.workspace_path / 'uploads'
    if not uploads.exists():
        uploads.mkdir(parents=True)
    path = uploads / filename
    file.save(str(path))
    log.info(f"[UPLOAD] Uploading file to {filename} to {path}")
    return path


def upload_file_is_allowed(file):
    extension = Path(file.filename).suffix
    log.debug(f"[ZIP] Extension for {file.filename}: {extension}")
    if not extension == '.zip':
        raise errors.PortalAPIError(
            400,
            f"File with extension \"{extension}\" is not allowed: #{file.filename}"
        )


class SubmissionsService(object):
    def create_submission(self, user: User, project: Project, submission_params: dict):
        """Creates a new submission in the database and downloads its files.

          Zip and git sources are processed differently.

          Args:
              user(User): the submission's author
              project(Project): the project the submission falls under
              submission_params(dict): custom parameters specified by
                  project with their values as a JSON-encoded string

          Returns(Submission): the created submission entity

          """
        # adding review at submission create is not supported

        new_submission = Submission(
            user=user, project=project, parameters=submission_params)
        write_entity(new_submission)

        # download relevant files into Storage
        self._submission = new_submission
        self.process_new_submission()
        return new_submission

    def __init__(self, submission: Submission = None):
        self._submission = submission

    @property
    def submission(self) -> Submission:
        return self._submission

    @property
    def project(self) -> Project:
        return self.submission.project

    @property
    def storage(self):
        return storage

    @property
    def request(self):
        return flask.request

    @property
    def storage_submission(self):
        return storage.submissions.get(self.submission.id)

    def process_new_submission(self) -> AsyncResult:
        project = self.submission.project
        self.submission.parameters['file_params'] = project.config.file_whitelist
        write_entity(self.submission)
        task = self.create_process_submission_task()
        return task

    def create_process_submission_task(self) -> AsyncResult:
        result: AsyncResult = tasks.process_submission.delay(self.submission.id)
        self.submission.storage_task_id = result.task_id
        return result

    def copy_submission(self, note: str) -> Submission:
        """Copies a submission. Used at resubmitting

        Args:
            note(str): Note for the submission. Typically contains a reason for the resubmit.

        Returns(Submission): Copied submission
        """
        log.info(f"[COPY] Submission {self.submission.id} to "
                 f"same project {self.submission.project.id}")
        new_submission = Submission(user=self.submission.user, project=self.project,
                                    parameters=self.submission.parameters)
        new_submission.note = note
        write_entity(new_submission)
        tasks.clone_submission_files.delay(
            source_id=self.submission.id, target_id=new_submission.id)
        return new_submission

    def delete_submission(self):
        """Deletes a submission

        """
        log.info(f"[DELETE] Submission {self.submission}")
        # TODO: dispatch action to clean the storage
        delete_entity(self.submission)

    def update_submission_state(self, client: Union[User, Worker], data: dict) -> Submission:
        """Updates submission's state

        Args:
            client(User): Client attempting the update
            data(dict): should contain one key, 'state' with the new value

        Returns(Submission): Updated submission
        """
        new_state = data['state']
        if isinstance(client, User) and new_state != SubmissionState.CANCELLED:
            raise errors.ForbiddenError(client,
                                        note=f"User {client.id} cannot update "
                                             f"state to other than CANCELLED.")
        self.submission.change_state(new_state)
        write_entity(self.submission)
        log.info(f"[UPDATE] Submission state {self.submission.id} "
                 f"by {client.id} to {self.submission.state}")
        return self.submission

    def cancel_submission(self):
        """Cancels the submission

        """
        processor = submission_processor.SubmissionProcessor(self.submission)
        processor.revoke_task()

    def send_zip(self, storage_submission: entities.Submission):
        storage_submission = storage_submission or self.storage_submission
        path = storage_submission.zip_path
        if not path.exists():
            raise errors.PortalAPIError(400, f"Requested path does not exist: {path}")
        log.debug(
            f"[SEND] Sending zip file for submission ({storage_submission.entity_id}): {path}")
        return flask.send_file(str(path), attachment_filename=path.name)

    def send_selected_file(self, storage_entity: entities.Submission, path_query: str):
        storage_entity = storage_entity or self.storage_submission
        path = storage_entity.get(path_query)
        log.debug(f"[SEND] Sending file for submission ({storage_entity.entity_id}): {path}")
        return flask.send_file(str(path), attachment_filename=path.name)

    def send_file_or_zip(self, storage_entity=None):
        storage_entity = storage_entity or self.storage_submission
        path_query = self.request.args.get('path')
        if path_query is None:
            return self.send_zip(storage_entity)
        return self.send_selected_file(storage_entity, path_query)

    def send_files_tree(self, storage_entity=None):
        storage_entity = storage_entity or self.storage_submission
        tree = storage_entity.tree()
        log.debug(f"[TREE] Tree for the storage entity {storage_entity.entity_id}: {tree} ")
        return tree

    def upload_results_to_storage(self):
        path = self.get_upload_file_path()
        task = tasks.upload_results_to_storage.delay(self.submission.id, path=str(path))
        return task

    def get_upload_file_path(self):
        file = self.request.files['file']
        upload_file_is_allowed(file)
        path = upload_files_to_storage(file)
        return path
